#include <iostream>
#include <sstream>

#include "assert.h"

#include "compile.hh"
#include "symtables.hh"
#include "binding.hh"
#include "basfun.hh"
#include "tristate"

#define CPP_SUBFUNC_NAME(x) (std::string(x)+"_f")
#define CPP_MEMBER_NAME(x) (std::string(x))
#define CPP_UTYPE_TYPENAME(x)  (std::string(x)+"_ut")
#define CPP_UTYPE_NAME(x)  (std::string(x)+"_u")
#define CPP_INT_NAME(x)    (std::string(x)+"_i")
#define CPP_SINGLE_NAME(x) (std::string(x)+"_z")
#define CPP_DOUBLE_NAME(x) (std::string(x)+"_d")
#define CPP_LONG_NAME(x)   (std::string(x)+"_l")
#define CPP_STRING_NAME(x) (std::string(x)+"_s")
#define CPP_ARRAY_NAME(x)  (std::string(x)+"_a")

static bool DoResumeHandling = false; /* FIXME: Make a commandline option or something */

std::vector<std::string> global_data;

static std::string FunctionReturnVarName;

static const std::pair<std::string, std::string>
    GetTypeBaseNames(const TypeInfo& type, const std::string& name)
{
    switch(type.Type)
    {
        case TypeInfo::Integer: return std::make_pair("BAS_INT", CPP_INT_NAME(name));
        case TypeInfo::Single:  return std::make_pair("BAS_SGN", CPP_SINGLE_NAME(name));
        case TypeInfo::Double:  return std::make_pair("BAS_DBL", CPP_DOUBLE_NAME(name));
        case TypeInfo::Long:    return std::make_pair("BAS_LNG", CPP_LONG_NAME(name));
        case TypeInfo::String:
            return type.fixlength
                ? /* FIXME */
                  std::make_pair("BAS_STR", CPP_STRING_NAME(name))
                : std::make_pair("BAS_STR", CPP_STRING_NAME(name));
        case TypeInfo::UserType:
            return std::make_pair(CPP_UTYPE_TYPENAME(type.utypename), CPP_UTYPE_NAME(name));
    }
    /* not reached, but keeps compiler happy */
    assert(!"not_reached");
    return std::make_pair(CPP_UTYPE_TYPENAME(type.utypename), CPP_UTYPE_NAME(name));
}

enum StrCompileDefs
{
    StrCompileMultipurpose,
    StrCompilePrinting,
    StrCompileConstOnly
};
static const std::string CompileStringConst
   (const std::string&str,
    StrCompileDefs defs = StrCompileMultipurpose)
{
    char esc = '"';
    if(defs == StrCompilePrinting && str.size() == 1)
    {
        esc = '\'';
    }
    std::string result;
    result += esc;
    bool contains_nul_char = false;
    unsigned size_minus = 0;
    for(unsigned a=0; a<str.size(); ++a)
    {
        unsigned char c = str[a];
        if((result.size()-size_minus) >= 77 && esc == '"')
        {
            result += "\"\n\"";
            size_minus = result.size()-1;
        }
        if(c=='\n') { result += "\\n"; continue; }
        if(c=='\r') { result += "\\r"; continue; }
        if(c=='\t') { result += "\\t"; continue; }
        if(c=='\b') { result += "\\b"; continue; }
        if(c=='\a') { result += "\\a"; continue; }
        if(c==esc || c=='\\') { result += '\\'; result += str[a]; continue; }
        if(c < 32 || c >= 127)
        {
            if(c == 0) contains_nul_char = true;
            char Buf[64];
            int width = 1;
            if(a+1 < str.size() && str[a+1]>='0' && str[a+1]<='7') width=3;
            sprintf(Buf, "\\%0*o", width, c);
            result += Buf;
            continue;
        }
        result += str[a];
    }
    result += esc;
    if(esc == '"' && defs != StrCompileConstOnly && contains_nul_char)
    {
        std::stringstream res2;
        res2 << "std::string(" << result << ", " << str.size() << ")";
        result = res2.str();
    }
    return result;
}

enum VarDeclType { Init_Var, Param_Ref, Param_Copy, Param_Const };

static struct
{
    std::string resulting_name;
    
    void operator() (std::ostream& out, const std::string& name,
                     TypeP typep, VarDeclType decl_type = Init_Var
                    )
    {
        assert(typep);
        const TypeInfo& type = *typep;

        const std::pair<std::string, std::string> base = GetTypeBaseNames(type, name);

        switch(type.arraytype)
        {
            case TypeInfo::NotArray:
            {
                bool const_ref = false;
                if(decl_type == Param_Const)
                    const_ref = !type.is_num();

                if(const_ref) out << "const ";
                out << base.first << ' ';
                if(decl_type == Param_Ref || const_ref) out << "&";
                out << (resulting_name=base.second);
                break;
            }
            case TypeInfo::StaticArray:
            {
                out << base.first << ' ';
                if(decl_type == Param_Copy)
                {
                    fprintf(stderr, "Warning: BYVAL Array parameters not supported\n");
                }
                out << (resulting_name=CPP_ARRAY_NAME(base.second));
                for(unsigned a=0; a<type.dimensions.size(); ++a)
                    out << (!a ? "[" : "][") << type.dimensions[a].second;
                out << "]";
                break;
            }
            case TypeInfo::DynamicArray:
            {
                if(decl_type == Param_Const) out << "const ";
                out << "BAS_DYNARRAY<" << base.first << ", " << type.dimensions.size() << "> ";
                if(decl_type == Param_Ref
                || decl_type == Param_Const) out << '&';
                out << (resulting_name=CPP_ARRAY_NAME(base.second));
                
                if(decl_type == Init_Var)
                {
                    out << "(empty_struct()";
                    for(unsigned a=0; a<type.dimensions.size(); ++a)
                    {
                        out << ", "
                            << type.dimensions[a].first
                            << ','
                            << type.dimensions[a].second;
                    }
                    out << ')';
                }
                break;
            }
            case TypeInfo::ParamArray:
            {
                if(decl_type == Param_Const) out << "const ";
                out << "BAS_MAGICARRAY<" << base.first << ">";
                if(decl_type == Param_Ref
                || decl_type == Param_Const) out << "&";
                out << (resulting_name=CPP_ARRAY_NAME(base.second));
                break;
            }
        }
    }
private:
} CompileVarDeclaration;

static void CompileTypeDefs(const Scope& scope)
{
    for(hash_map<std::string, UtypeP>::const_iterator
        i = scope.types.begin();
        i != scope.types.end();
        ++i)
    {
        UtypeP p = i->second;
        assert(p);
        
        std::cout << "struct " << CPP_UTYPE_TYPENAME(i->first) << " {\n";
        
        for(unsigned a=0; a<p->size(); ++a)
        {
            std::cout << "  ";
            CompileVarDeclaration(std::cout, CPP_MEMBER_NAME((*p)[a].first), (*p)[a].second );
            std::cout << ";\n";
        }
        
        std::cout << "};\n";
    }
}

static void CompileVars(const ScopeVarListType& vars, tristate shared_only, bool implicit_static=false)
{
    for(ScopeVarListType::const_iterator i = vars.begin(); i != vars.end(); ++i)
    {
        ctxp ctx = i->second;
        assert(ctx);
        if(!shared_only.is_maybe() && ctx->is_shared != shared_only) continue;
        if(ctx->param_index >= 0) continue;

        if(ctx->is_static || implicit_static) std::cout << "static ";
        
        CompileVarDeclaration(std::cout, i->first, ctx->type);
        std::cout << ";\n";
    }
}

static void CompileVars(const Scope& scope, tristate shared_only, bool implicit_static=false)
{
    CompileVars(scope.Utype  .scalars, shared_only, implicit_static);
    CompileVars(scope.Utype  .arrays, shared_only, implicit_static);
    CompileVars(scope.Int    .scalars, shared_only, implicit_static);
    CompileVars(scope.Int    .arrays, shared_only, implicit_static);
    CompileVars(scope.Single .scalars, shared_only, implicit_static);
    CompileVars(scope.Single .arrays, shared_only, implicit_static);
    CompileVars(scope.Double .scalars, shared_only, implicit_static);
    CompileVars(scope.Double .arrays, shared_only, implicit_static);
    CompileVars(scope.Long   .scalars, shared_only, implicit_static);
    CompileVars(scope.Long   .arrays, shared_only, implicit_static);
    CompileVars(scope.String .scalars, shared_only, implicit_static);
    CompileVars(scope.String .arrays, shared_only, implicit_static);
}

static void CompileSubFunc(const std::string& name, const SubRoutine& sub, bool decl_only)
{
    std::cout << "static ";
    if(sub.returntype)
    {
        const std::pair<std::string, std::string> base
            = GetTypeBaseNames(*sub.returntype, name);
        /* no "const" here, to not hinder passing the result as reference */
        std::cout << base.first;
    }
    else
        std::cout << "void";

    std::cout << " " << CPP_SUBFUNC_NAME(name) << "(";
    
    for(unsigned a=0; a<sub.params.size(); ++a)
    {
        const SubRoutine::ParamSpec& param = sub.params[a];
        if(a) std::cout << ", ";
        CompileVarDeclaration(std::cout,
                              param.name,
                              param.type,
                              param.was_modified
                               ? param.passconvention ==
                                   SubRoutine::ParamSpec::ByVal
                                       ? Param_Copy
                                       : Param_Ref
                               : Param_Const);
    }
    
    std::cout << ')';
    
    if(decl_only)
        std::cout << ";\n";
    else
    {
        std::cout << "\n";
    }
}

static void CompileSubDefs(const Scope& scope)
{
    for(hash_map<std::string, SubFuncP>::const_iterator
        i = scope.subroutines.begin(); i != scope.subroutines.end(); ++i)
    {
        SubFuncP sub = i->second; assert(sub);
        CompileSubFunc(i->first, *sub, true);
    }
}

static void CompileCode(stmtp codep)
{
    std::list<stmtp> stack;
    while(codep) { stack.push_front(codep); codep=codep->get_prev(); }
    for(std::list<stmtp>::iterator
        j, i = stack.begin(); i != stack.end(); i=j)
    {
        j=i; ++j;
        codep = *i;
        
        if(DoResumeHandling && codep->MayErr()) std::cout << "B_ERROR_BEGIN() ";
        codep->Compile();
        if(DoResumeHandling && codep->MayErr()) std::cout << " B_ERROR_OVER() ";
        
        stack.erase(i);
    }
}

static const std::string GenerateLabelName(long double lab)
{
    const Scope& scope = *GetCurrentScope();
    if(scope.labels.find(lab) == scope.labels.end())
    {
        fprintf(stderr, "Error: Label not defined in current scope\n");
        return "L??";
    }
    unsigned lnum = 0;
    static std::map<std::pair<const Scope*, long double>, unsigned> ldata;
    std::pair<const Scope*, long double> seed(&scope, lab);
    std::map<std::pair<const Scope*, long double>, unsigned>::const_iterator
        l = ldata.find(seed);
    if(l != ldata.end()) lnum = l->second;
    else lnum = ldata[seed] = ldata.size() + 1;
    std::stringstream tmp; tmp << 'L' << lnum;
    return tmp.str();
}

static void CompileExpr(exprp e,
                        ExprCompileDefs defs=exprdef_nothingparticular)
{
    if(!e) return;
    
    if(defs == exprdef_cast_to_int)
    {
        TypeP type = e->get_type();
        if(type && type->is_num())
        {
            if(type->Type != TypeInfo::Integer
            && type->Type != TypeInfo::Long)
            {
                if(expr_number* num = dynamic_cast<expr_number*>(&*e))
                {
                    if(num->is_int()) goto NormalExprCompile;
                }
                
                std::cout << "BasCLNG(";
                defs = exprdef_nothingparticular;
                e->Compile(defs);
                std::cout << ")";
                return;
            }
        }
    }
NormalExprCompile:
    e->Compile(defs);
}

static void VerifyLabels(const Scope& scope)
{
    for(std::map<long double, LabelDef>::const_iterator
        i = scope.labels.begin(); i != scope.labels.end(); ++i)
    {
        if(!i->second.defined)
            fprintf(stderr, "Error: Label %Lg not defined\n", i->first);
    }
}

/*
PLAN:
    Define typedefs. - DONE
    Define global variables. -SHARED ONLY DONE
    Declare subroutines. -DONE
    
    Compile code.
    
    Compile each function.

NOTES:
    Constants are not defined. They are compiled!

TODO:
    SHARED imports?
    
*/
void Compile(stmtp main_codep,
             const std::vector<std::string>& global_data)
{
    ScopeP scope = GetRootScope();
    SetCurrentScope(scope);
    
    VerifyLabels(*scope);
    AnalyzeSubParamUsage();
    
    std::cout << "#include <qblib/qblib.hh>\n";
    
    std::cout << "static const unsigned DATA_COUNT = " << global_data.size() << ";\n";
    if(!global_data.empty())
    {
        std::cout << "static unsigned DATA_PTR = 0;\n";
        std::cout << "static const char* const DATA_TABLE[DATA_COUNT] = {\n";
        for(unsigned a=0; a<global_data.size(); ++a)
        {
            //std::cout << "// " << a << "\n";
            std::cout << CompileStringConst(global_data[a], StrCompileConstOnly);
            if(a+1 < global_data.size()) std::cout << ',';
            std::cout << "\n";
        }
        std::cout << "}\n";
    }
    
    CompileTypeDefs(*scope);
    CompileVars(*scope, true, true); // shared only
    CompileSubDefs(*scope);

    std::cout << "int main(int argc, char**argv)\n";
    std::cout << "{\n";
    CompileVars(*scope, false, false); // local only
    std::cout << "B_ERR_HEADER();\n";
    CompileCode(main_codep);
    std::cout << "}\n";

    /* Compile subs */
    for(hash_map<std::string, SubFuncP>::const_iterator
        i = scope->subroutines.begin(); i != scope->subroutines.end(); ++i)
    {
        SubFuncP sub = i->second; assert(sub);
        ScopeP subscope = sub->sub_scope;
        SetCurrentScope(subscope);
        
        VerifyLabels(*subscope);
        
        std::cout << "\n\n";
        
        CompileSubFunc(i->first, *sub, false);
        std::cout << "{\n";
        
        if(sub->returntype)
        {
            CompileVarDeclaration(std::cout, "FunctionReturn", sub->returntype);
            FunctionReturnVarName = CompileVarDeclaration.resulting_name;
            std::cout << ";\n";
        }
        
        CompileVars(*subscope, false, sub->is_static); // local only
        CompileCode(sub->code);
        
        std::cout << "subfunc_exit:; \n";

        if(sub->returntype)
        {
            std::cout << "    return " << FunctionReturnVarName << ";\n";
        }
        std::cout << "}\n";
    }
}

void stmt_if::Compile()
{
    if(!thenp && elsep && elseifp.empty())
    {
        std::cout << "if(!("; CompileExpr(condition, exprdef_pure_condition);
        std::cout << ")) {\n"; CompileCode(elsep);
        std::cout << "}\n";
    }
    else
    {
        std::cout << "if("; CompileExpr(condition, exprdef_pure_condition);
        std::cout << ") {\n";
            if(thenp) CompileCode(thenp);
        std::cout << "}";
        for(std::list<std::pair<exprp, stmtp> >::const_iterator
            i = elseifp.begin(); i != elseifp.end(); ++i)
        {
            std::cout << " else if("; CompileExpr(i->first, exprdef_pure_condition);
            std::cout << ") {\n";
            if(i->second) CompileCode(i->second);
            std::cout << "}";
        }
        if(elsep)
        {
            std::cout << " else {\n";
            CompileCode(elsep);
            std::cout << "}";
        }
        std::cout << "\n";
    }
}
void stmt_doloop::Compile()
{
    std::cout << "{ __label__ do_exit;\n";
    if(precondition)
    {
        std::cout << "while("; CompileExpr(precondition, exprdef_pure_condition); std::cout << ")\n{\n";
    }
    else if(!postcondition)
        std::cout << "for(;;)\n{\n";
    else
        std::cout << "do {\n";

    CompileCode(bodyp);
    
    if(postcondition)
    {
        if(precondition)
        {
            std::cout << "if("; CompileExpr(postcondition, exprdef_pure_condition);
            std::cout << ") break;";
        }
        else
        {
            std::cout << "} while("; CompileExpr(postcondition, exprdef_pure_condition);
            std::cout << ");";
        }
    }
    else
        std::cout << "}";
    std::cout << " do_exit:; }\n";
}
void stmt_for::Compile()
{
    static unsigned ForCounter=0;
    char ForTempName[64]; sprintf(ForTempName, "FORTEMP%u", ForCounter++);
    
    bool nonconst_step = true;
    QBtoken step_operator = (QBtoken)'<';
    
    if(!step)
        nonconst_step = false;
    else
    {
        if(expr_number* num = dynamic_cast<expr_number*> (&*step))
        {
            nonconst_step = false;
            if(num < 0) step_operator = (QBtoken)'>';
        }
    }
    
    std::pair<std::string, std::string> ForTemp;
    if(nonconst_step)
    {
        ForTemp = GetTypeBaseNames(*step->get_type(), ForTempName);
    }
    
    std::cout << "{ __label__ for_exit;\n";
    if(nonconst_step)
    {
        std::cout << "const " << ForTemp.first << " " << ForTempName << " = ";
        CompileExpr(step);
        std::cout << ";\n";
    }
    std::cout << "for(";
        stmt_let(var, begin).Compile(); // it appends a semicolon, too.
        if(nonconst_step)
        {
            std::cout << "(" << ForTempName << " < 0 ? (";
            //negative step?
                CompileExpr(var); std::cout << " > "; CompileExpr(end);
            std::cout << ") : (";
            //positive step?
                CompileExpr(var); std::cout << " < "; CompileExpr(end);
            std::cout << ')';
        }
        else
        {
            expr_binary(step_operator, var, end).Compile(exprdef_pure_condition);
        }
        
        std::cout << "; ";
        
        CompileExpr(var);
        if(!step)
            { std::cout << " += 1 "; }
        else if(nonconst_step)
            { std::cout << " += " << ForTempName; }
        else
            { std::cout << " += "; CompileExpr(step); }
    std::cout << ")\n{\n";
        CompileCode(bodyp);
    std::cout << "} for_exit:; }\n";
}
void stmt_let::Compile()
{
    CompileExpr(target);
    
    if(expr_binary* bin = dynamic_cast<expr_binary*> (&*value))
    {
        std::string op, cast1;
        ExprCompileDefs subdefs = exprdef_nothingparticular;
        switch(bin->token)
        {
            case XOR: op = "^="; break;
            case OR:  op = "|="; break;
            case AND: op = "&="; break;
            case (QBtoken)'+': op = "+="; break;
            case (QBtoken)'-': op = "-="; break;
            case (QBtoken)'*': op = "*="; break;
            case (QBtoken)'/': op = "/="; cast1 = "(BAS_DBL)"; break;
            default: break;
        }
        if(!op.empty())
        {
            /* Check if the first argument in this expr_binary
             * is the same as target, and if it is, act on it
             */
            if(expr_var* tvar = dynamic_cast<expr_var*>(&*target))
                if(expr_var* svar = dynamic_cast<expr_var*>(&*bin->p0))
                {
                    if(tvar->getname() == svar->getname()
                    && *tvar->get_type() == *svar->get_type()
                    && !tvar->isarray() && !svar->isarray())
                    {
                        std::cout << ' ' << op << ' ' << cast1;
                        CompileExpr(bin->p1, subdefs);
                        std::cout << ";\n";
                        return;
                    }
                }
        }
    }
    
    std::cout << " = ";
    CompileExpr(value);
    std::cout << ";\n";
}

static void CompileCaseTest(exprp expr, const std::string& varname,
                            bool IsSwitchCase)
{
    ExprCompileDefs defs = IsSwitchCase ? exprdef_cast_to_int : exprdef_nothingparticular;
    if(expr_caseor* e = dynamic_cast<expr_caseor*>(&*expr))
    {
        CompileCaseTest(e->a, varname, defs);
        if(!IsSwitchCase) std::cout << " || ";
        CompileCaseTest(e->b, varname, defs);
    }
    else if(expr_caseis* e = dynamic_cast<expr_caseis*>(&*expr))
    {
        std::string op = "==";
        switch(e->token)
        {
            case (QBtoken)'>':    op = ">"; break;
            case (QBtoken)'<':    op = "<"; break;
            case (QBtoken)'=':    op = "=="; break;
            case NOTEQUAL:        op = "!="; break;
            case LESSOREQUAL:     op = "<="; break;
            case GREATEROREQUAL:  op = ">="; break;
            default: assert(!"unexpected caseis token");
        }
        if(IsSwitchCase && op == "==") op="";
        if(IsSwitchCase) std::cout << "case " << op;
        else std::cout << varname << ' ' << op << ' ';
        CompileExpr(e->ref, defs);
        if(IsSwitchCase) std::cout << ":\n";
    }
    else if(expr_caserange* e = dynamic_cast<expr_caserange*>(&*expr))
    {
        if(IsSwitchCase)
        {
            std::cout << "case ";
            CompileExpr(e->min, defs);
            std::cout << " ... ";
            CompileExpr(e->max, defs);
            std::cout << ":\n";
        }
        else
        {
            std::cout << "(" << varname << " >= ";
            CompileExpr(e->min, defs);
            std::cout << " && " << varname << " <= ";
            CompileExpr(e->max, defs);
            std::cout << ')';
        }
    }
    else assert(!"unexpected case type");
}

void stmt_select::Compile()
{
    if(IsEntirelyIntegerCase())
    {
        std::cout << "switch(";
        CompileExpr(ref, exprdef_cast_to_int);
        std::cout << ")\n{\n";
        for(std::list<
            std::pair<exprp, stmtp>
                 >::const_iterator i = cases.begin(); i != cases.end(); ++i)
        {
            if(dynamic_cast<expr_caseelse*>(&*i->first))
            {
                std::cout << "default:\n";
                CompileCode(i->second);
                break;
            }
            CompileCaseTest(i->first, "?", true);
            CompileCode(i->second);
            std::cout << "break;\n";
        }
        std::cout << "}\n";
    }
    else
    {
        static unsigned SwitchCounter=0;
        char SwitchTempName[64]; sprintf(SwitchTempName, "SWITCHTEMP%u", SwitchCounter++);
        std::pair<std::string, std::string> SwitchTemp;
        SwitchTemp = GetTypeBaseNames(*ref->get_type(), SwitchTempName);
        
        bool first = true;
        std::cout << "{ const " << SwitchTemp.first << " " << SwitchTemp.second
                  << " = "; CompileExpr(ref);
        std::cout << ";\n";
        
        for(std::list<
            std::pair<exprp, stmtp>
                 >::const_iterator i = cases.begin(); i != cases.end(); ++i)
        {
            if(!first) std::cout << "else ";
            if(dynamic_cast<expr_caseelse*>(&*i->first))
            {
                std::cout << "{\n";
                CompileCode(i->second);
                std::cout << "} ";
                break;
            }
            std::cout << "if(";
            CompileCaseTest(i->first, SwitchTemp.second, false);
            std::cout << ")\n{\n";
            CompileCode(i->second);
            std::cout << "} ";
            first=false;
        }
    
        std::cout << "}\n";
    }
}

void stmt_print::Compile()
{
    static unsigned UsingCounter=0;
    char TempName[64]; sprintf(TempName, "PTEMP%u", UsingCounter++);
    std::string OutTempName="OUT"+std::string(TempName),
              UsingTempName="USING"+std::string(TempName);

    std::cout << "{ PrintStreamBase& " << OutTempName << " = ";
    if(target)
    {
        std::cout << "GET_STREAM_OUTPUT("; CompileExpr(target);
        std::cout << ')';
    }
    else if(is_lprint)
    {
        std::cout << "GET_PRINTER()";
    }
    else
    {
        std::cout << "GET_STDOUT()";
    }
    std::cout << ";\n";
    
    if(using_string)
    {
        std::cout << "UsingFormatter " << UsingTempName << "(";
        CompileExpr(using_string);
        std::cout << ");\n";
    }
    
    bool outputting = false;
    
    for(std::list<std::pair<QBtoken, exprp> >::const_iterator
        i = params.begin(); i != params.end(); ++i)
    {
        switch(i->first)
        {
            case (QBtoken)0:
                if(using_string)
                {
                    if(outputting) { std::cout << ";\n"; outputting=false; }
                    std::cout << UsingTempName << ".DoParam(" << OutTempName
                              << ", ";
                    CompileExpr(i->second);
                    std::cout << ");\n";
                }
                else
                {
                    if(!outputting) { outputting=true; std::cout << OutTempName; }
                    std::cout << " << ";
                    CompileExpr(i->second, exprdef_may_produce_char_from_string);
                }
                break;
            case SPC:
                if(outputting) { std::cout << ";\n"; outputting=false; }
                std::cout << OutTempName << ".DoSPC("; CompileExpr(i->second);
                std::cout << ");\n";
                break;
            case TAB:
                if(outputting) { std::cout << ";\n"; outputting=false; }
                if(i->second)
                {
                    std::cout << OutTempName << ".DoTAB("; CompileExpr(i->second);
                    std::cout << ");\n";
                }
                else
                {
                    std::cout << OutTempName << ".DoNextTAB();\n";
                }
                break;
            case CRLF:
                if(outputting) { std::cout << ";\n"; outputting=false; }
                std::cout << OutTempName << ".DoCRLF();\n";
                break;
            default: assert(!"unexpected print operator");
        }
    }
    
    if(outputting) { std::cout << "; "; outputting=false; }
    std::cout << "}\n";
}

void stmt_input::Compile()
{
    static unsigned UsingCounter=0;
    char TempName[64]; sprintf(TempName, "PTEMP%u", UsingCounter++);
    std::string InTempName="IN"+std::string(TempName);

    std::cout << "{ InputStreamBase& " << InTempName << " = ";
    if(source)
    {
        std::cout << "GET_STREAM_INPUT("; CompileExpr(source);
        std::cout << ')';
    }
    else
    {
        std::cout << "GET_STDIN()";
    }
    std::cout << ";\n";

    if(!is_linput && !source && !prompt) prompt = new expr_string("? ");
    
    if(!source && !is_linput)  std::cout << "for(;;) {";
    
    if(!source && prompt)
        { std::cout << "GET_STDOUT() << ";
          CompileExpr(prompt, exprdef_may_produce_char_from_string);
          std::cout << ";\n"; }
    
    for(std::list<exprp>::const_iterator
        i = params.begin(); i != params.end(); ++i)
    {
        std::string method = is_linput ? "ReadLine" : "InputTo";
        std::cout << InTempName << '.' << method
                  << '('; CompileExpr(*i); std::cout << ");\n";
    }
    
    /* Input error from console results in "?Redo from start"
     * Input error from file results in error 62 */
    
    if(!source && !is_linput)
        std::cout << "if(" << InTempName << ".Satisfied()) break; \n"
                  << "GET_STDOUT() << " << InTempName << ".GetRedoMsg(); }\n";
    else
        std::cout << "if(!" << InTempName << ".Satisfied()) B_ERROR(62);\n";
    
    if(!source && !keep_cursor)
        std::cout << "GET_STDOUT().DoCRLF();\n";
        
    std::cout << "}\n";
}


void stmt_labelref::Compile()
{
    expr_label* e = 0;
    long double lab=0;
    std::string lname;
    if(label)
    {
        e = dynamic_cast<expr_label*> (&*label);
        assert(e);
        lab = e->get_label();
        lname = GenerateLabelName(lab);
    }
    switch(token)
    {
        case EXITDO: std::cout << "goto do_exit;\n"; break;
        case EXITFOR: std::cout << "goto for_exit;\n"; break;
        case END: std::cout << "B_END();\n"; break;
        case STOP: std::cout << "B_STOP();\n"; break;
        case EXITSUB:
        case EXITFUNCTION:
        case EXITDEF: std::cout << "goto subfunc_exit;\n"; break;
        case RETURN:
            if(label)
                std::cout << "B_RETURN_TO(" << lname << ");\n";
            else
                std::cout << "B_RETURN();\n";
            break;
        case RESUME:
            if(label)
                std::cout << "B_RESUME_TO(" << lname << ");\n";
            else
                std::cout << "B_RESUME();\n";
            break;
        case RESUMENEXT:
            std::cout << "B_RESUME_NEXT();\n";
            break;
        case GOTO:
            std::cout << "goto " << lname << ";\n";
            break;
        case GOSUB:
            std::cout << "B_GOSUB(" << lname << ");\n";
            break;
        case ON_ERROR:
            if(label)
                std::cout << "ON_ERROR_GOTO(" << lname << ");\n";
            else
                std::cout << "ON_ERROR_GOTO_NOWHERE();\n";
            break;
        case ON_PLAY: /* FIXME, not considered */
            if(label)
                std::cout << "ON_PLAY_GOTO(" << lname << ");\n";
            else
                std::cout << "ON_PLAY_GOTO_NOWHERE();\n";
            break;
        case ON_KEY: /* FIXME, not considered */
            if(label)
                std::cout << "ON_KEY_GOTO(" << lname << ");\n";
            else
                std::cout << "ON_KEY_GOTO_NOWHERE();\n";
            break;
        case ON_TIMER: /* FIXME, not considered */
            if(label)
                std::cout << "ON_TIMER_GOTO(" << lname << ");\n";
            else
                std::cout << "ON_TIMER_GOTO_NOWHERE();\n";
            break;
        case RESTORE:
            std::cout << "B_RESTORE(";
            if(!label) std::cout << '0';
            else
            {
                const Scope& scope = *GetCurrentScope();
                std::map<long double, LabelDef>::const_iterator
                    i = scope.labels.find(lab);
                if(i != scope.labels.end())
                    std::cout << i->second.dataptr;
                
                std::cout << " /* " << lname << " */";
            }
            std::cout << ");\n";
            break;
        default: assert(!"unexpected stmt_goto token");
    }
}
void stmt_label::Compile()
{
    long double lab = label;
    std::string lname = GenerateLabelName(lab);
    std::cout << lname << ": ;\n";
}

void stmt_fileopen::Compile() { stmt::Compile(); }

void stmt_subcall::Compile()
{
    exprp e = new expr_userfun(name, 0, params);
    CompileExpr(e);
    std::cout << ";\n";
}
void stmt_read::Compile()
{
    std::cout << "B_READ("; CompileExpr(target);
    std::cout << ");\n";
}

void stmt_redim::Compile()
{
    expr_var* v = dynamic_cast<expr_var*>(&*target);
    assert(v);
    const std::string& name = v->getname();
    
    VarContext ctx = FindArrayType(name);
    if(!ctx) fprintf(stderr, "Cannot find definition of '%s'\n", name.c_str());
    TypeP type = ctx.info->type;

    std::pair<std::string, std::string>
        base = GetTypeBaseNames(*type, BasUCASE(StripTypeSuffix(name)));
    
    switch(type->arraytype)
    {
        case TypeInfo::NotArray:
        case TypeInfo::StaticArray: std::cout << "/* REDIM eaten by grue? */\n"; break;
        case TypeInfo::DynamicArray:
        case TypeInfo::ParamArray:
            std::cout << CPP_ARRAY_NAME(base.second);
            std::cout << ".Redim(" << dimensions.size();
            for(unsigned a=0; a<dimensions.size(); ++a)
            {
                std::cout << ", ";
                CompileExpr(dimensions[a].first, exprdef_cast_to_int);
                std::cout << ',';
                CompileExpr(dimensions[a].second, exprdef_cast_to_int);
            }
            std::cout << ");\n";
            break;
    }
}
void stmt_erase::Compile()
{
    expr_var* v = dynamic_cast<expr_var*>(&*target);
    assert(v);
    const std::string& name = v->getname();
    std::pair<std::string, std::string>
        base = GetTypeBaseNames(*v->get_type(), BasUCASE(StripTypeSuffix(name)));
    VarContext ctx = FindArrayType(name);
    if(!ctx) fprintf(stderr, "Cannot find definition of '%s'\n", name.c_str());
    TypeP type = ctx.info->type;
    switch(type->arraytype)
    {
        case TypeInfo::NotArray: std::cout << "/* ERASE eaten by grue? */\n"; break;
        case TypeInfo::StaticArray:
        {
            for(unsigned a=0; a<type->dimensions.size(); ++a)
                std::cout << "for(unsigned EraTmp" << a << "=0; EraTmp" << a
                          << "<"
                          << type->dimensions[a].second
                          << "; ++EraTmp" << a
                          << ")\n";
            std::cout << CPP_ARRAY_NAME(base.second);
            for(unsigned a=0; a<type->dimensions.size(); ++a)
                std::cout << "[EraTmp" << a << ']';
            std::cout << " = " << base.first << "();\n";
            break;
        }
        case TypeInfo::DynamicArray:
        case TypeInfo::ParamArray:
            std::cout << CPP_ARRAY_NAME(base.second);
            std::cout << ".Erase();\n";
            break;
    }
}

static void CompileFunctionCall
    (const std::string& functionname,
     const std::vector<exprp>& params,
     const std::vector<SubRoutine::ParamSpec>& paramdefs
       /* Paramdefs components used:
           * type->arraytype == ParamArray
           * passconvention != ByVal
             && was_modified (if needs to be passed by reference)
           * type (for temporary VarDeclaration)
        */
    )
{
    std::vector<std::pair<bool, std::string> > temps(params.size());
    bool has_temps = false;
    
    for(unsigned a=0; a<params.size(); ++a)
    {
        temps[a].first = false;
        
        /* If the parameter needs to be passed by reference */
        if(paramdefs[a].type->arraytype == TypeInfo::ParamArray)
        {
            if(expr_var* v = dynamic_cast<expr_var*>(&*params[a]))
                if(v->isarray())
                {
                    const TypeInfo& vtype = *v->get_type();

                    if(vtype.arraytype == TypeInfo::ParamArray)
                    {
                        /* Param array does not need special care
                         * when the source is also a param array
                         */
                        continue;
                    }
                    
                    /* It's not a var, so it's a temporary then. */
                    if(!has_temps)
                    {
                        has_temps = true;
                        std::cout << "({"; // gcc extension: statement expressions
                    }
                    static unsigned RefCounter=0;
                    char TempName[64]; sprintf(TempName, "ArrTmp%u", RefCounter++);

                    CompileVarDeclaration(std::cout, TempName,
                                          paramdefs[a].type,
                                          Param_Copy);
                    temps[a].first  = true;
                    temps[a].second = CompileVarDeclaration.resulting_name;
                    
                    std::cout << '(';
                    
                    switch(vtype.arraytype)
                    {
                        case TypeInfo::NotArray: goto eek;
                        case TypeInfo::StaticArray:
                        {
                            std::cout << "&";
                            CompileExpr(params[a], exprdef_nothingparticular);
                            for(unsigned a=0; a<vtype.dimensions.size(); ++a)
                                std::cout << "[0]";
                            
                            std::cout << ", " << vtype.dimensions.size();
                            for(unsigned a=0; a<vtype.dimensions.size(); ++a)
                                std::cout << ", " << vtype.dimensions[a].first
                                          << ','  << vtype.dimensions[a].second;
                            break;
                        }
                        case TypeInfo::DynamicArray:
                        case TypeInfo::ParamArray:
                            CompileExpr(params[a], exprdef_nothingparticular);
                            break;
                    }
                    std::cout << ')';
                    
                    continue;
                }
        eek:
            fprintf(stderr, "Error: Attempt to pass a non-array to a function that expects array\n");
        }
        
        if(paramdefs[a].passconvention != SubRoutine::ParamSpec::ByVal
        && paramdefs[a].was_modified)
        {
            /* Check if the parameter to be passed is a temporary. */
            /* If it is, a temporary holder needs to be constructed. */
            if(!dynamic_cast<expr_var*>(&*params[a]))
            {
                /* It's not a var, so it's a temporary then. */
                if(!has_temps)
                {
                    has_temps = true;
                    std::cout << "({"; // gcc extension: statement expressions
                }
                
                static unsigned RefCounter=0;
                char TempName[64]; sprintf(TempName, "RefTmp%u", RefCounter++);

                CompileVarDeclaration(std::cout, TempName,
                                      paramdefs[a].type,
                                      Param_Copy);
                temps[a].first  = true;
                temps[a].second = CompileVarDeclaration.resulting_name;
                std::cout << '=';
                CompileExpr(params[a], exprdef_nothingparticular);
                std::cout << ';';
            }
        }
    }
    
    std::cout << functionname << '(';
    for(unsigned a=0; a<params.size(); ++a)
    {
        if(a) std::cout << ", ";
        
        if(temps[a].first)
            std::cout << temps[a].second;
        else
            CompileExpr(params[a], exprdef_nothingparticular);
    }
    std::cout << ')';
    
    if(has_temps)
    {
        std::cout << "})";
    }
}

static const std::vector<SubRoutine::ParamSpec>
    CreateParamDefsFromPattern(const std::string& pattern,
        std::vector<exprp>& params)
{
    std::vector<SubRoutine::ParamSpec> paramdefs(params.size());
    
    unsigned paramno=0;
    const char* defs = pattern.c_str();
    while(*defs)
    {
        #define EP() do { \
            while(paramno >= params.size()) params.push_back((exprp)0); \
            if(paramno >= paramdefs.size()) paramdefs.resize(paramno+1); } while(0)
        #define PTYPE() do { \
            if(params[paramno]) \
                paramdefs[paramno].type = params[paramno]->get_type(); \
            else \
                fprintf(stderr, "param %u is empty\n", paramno); \
            } while(0)
        #define SetByVal() do { \
            paramdefs[paramno].was_modified = false; \
            paramdefs[paramno].passconvention = SubRoutine::ParamSpec::ByVal; } while(0)
        #define SetByRef() do { \
            paramdefs[paramno].was_modified = true; \
            paramdefs[paramno].passconvention = SubRoutine::ParamSpec::ByRef; } while(0)
        #define InsInt(v) params.insert(params.begin()+paramno, new expr_number(v))
        #define InsStr(s) params.insert(params.begin()+paramno, new expr_string(s))
        
        if(paramno+1 > params.size()) params.push_back((exprp)0);
        switch(*defs++)
        {
            case 'n': case 'x':
                EP(); if(!params[paramno]) { params[paramno]=new expr_number(-1); }
                //passthru
            case 'N': case 'S': case 'X': // number or string
              GotS:
                EP(); PTYPE(); SetByVal(); ++paramno;
                break;
            case 's':
                EP(); if(!params[paramno]) { params[paramno]=new expr_string(""); }
                goto GotS;
            case 'p':
                EP(); if(!params[paramno]) { params[paramno]=new expr_number(0);
                                             InsInt(0); InsInt(-1); }
                //passthru
            case 'P': // step coordinate
                EP(); PTYPE(); SetByVal(); ++paramno;
                EP(); PTYPE(); SetByVal(); ++paramno;
                EP(); PTYPE(); SetByVal(); ++paramno;
                break;
            case 'c':
                EP(); if(!params[paramno]) { params[paramno]=new expr_number(0);
                                             InsInt(-1); }
                //passthru
            case 'C': // coordinate
                EP(); PTYPE(); SetByVal(); ++paramno;
                EP(); PTYPE(); SetByVal(); ++paramno;
                break;
            case 'V': // lvalue
                EP(); PTYPE(); SetByRef(); ++paramno;
                break;
            case 'A': // array target (GET, PUT)
                EP(); PTYPE(); SetByRef(); ++paramno;
                break;
            case 'L': // label
            case 'l':
                assert(!"Label params should only occur in stmt_labelref");
            default:
                fprintf(stderr, "pattern '%s': unknown char '%c'\n", pattern.c_str(), *defs);
        }
        
        #undef EP
        #undef PTYPE
        #undef SetByVal
        #undef SetByRef
        #undef InsInt
        #undef InsStr
    }
    /*
    fprintf(stderr, "params.size=%u, paramno=%u, paramdefs.size=%u, pattern=%s\n",
        (unsigned)params.size(), paramno, (unsigned)paramdefs.size(), pattern.c_str());
    */
    return paramdefs;
}

void stmt_lib_stmt::Compile()
{
    const char* funcname = 0;
    #define tokcase(t, p) \
        case t: funcname = "B_Stmt"#t; goto DoParams;
    switch(token)
    {
/* #include "libdata-stmt-label.inc" - handled separately */
#include "libdata-stmt-stmt.inc"

        default: assert(!"unknown stmt_lib_stmt token");
    }
    return;
DoParams:
    std::vector<exprp> plist = params;
    std::vector<SubRoutine::ParamSpec> paramdefs
        = CreateParamDefsFromPattern(pattern, plist);
    CompileFunctionCall(
        funcname,
        plist,
        paramdefs
    );
    #undef tokcase
    std::cout << ";\n";
}

void expr_string::Compile(ExprCompileDefs defs)
{
    std::cout << CompileStringConst
        (this->value,
           defs==exprdef_may_produce_char_from_string
           ? StrCompilePrinting
           : StrCompileMultipurpose);
}
void expr_number::Compile(ExprCompileDefs defs)
{
    if(is_int())
        std::cout << get_i();
    else
    {
        std::cout.precision(20);
        std::cout << get_f();
    }
}
void expr_unary::Compile(ExprCompileDefs defs)
{
    std::cout << "(";
    ExprCompileDefs subdefs = exprdef_nothingparticular;
    switch(token)
    {
        case NOT:
            if(defs == exprdef_pure_condition) { std::cout << "!"; subdefs=defs; break; }
            std::cout << "~";
            subdefs = exprdef_cast_to_int;
            break;
        case (QBtoken)'-':
            std::cout << "-";
            break;
        default: assert(!"unexpected expr_unary token");
    }
    CompileExpr(p, subdefs);
    std::cout << ')';
}
void expr_binary::Compile(ExprCompileDefs defs)
{
    std::string fucast, funame, cast0, op, cast1, finish;
    ExprCompileDefs subdefs = exprdef_nothingparticular;

    switch(token)
    {
        case XOR:
            op = "^";
        FinishInteger:
            subdefs = exprdef_cast_to_int;
            break;
        case OR:
            if(defs == exprdef_pure_condition && IsComparison())
                { op = "||"; subdefs=defs; break; }
            op = "|"; goto FinishInteger;
        case AND:
            if(defs == exprdef_pure_condition && IsComparison())
                { op = "&&"; subdefs=defs; break; }
            op = "&"; goto FinishInteger;
        case (QBtoken)'>':
            op = ">";
        FinishCompare:
            if(defs == exprdef_pure_condition)
                subdefs = defs;
            else
            {
                fucast = "(";
                finish = ") ? -1 : 0";
            }
            break;
        case (QBtoken)'<':    op = "<"; goto FinishCompare;
        case (QBtoken)'=':    op = "=="; goto FinishCompare;
        case NOTEQUAL:        op = "!="; goto FinishCompare;
        case LESSOREQUAL:     op = "<="; goto FinishCompare;
        case GREATEROREQUAL:  op = ">="; goto FinishCompare;
        case (QBtoken)'+':    op = "+"; break;
        case (QBtoken)'-':    op = "-"; break;
        case (QBtoken)'\\':   op = "/"; subdefs = exprdef_cast_to_int; break;
        case (QBtoken)'*':    op = "*"; break;
        case (QBtoken)'/':    op = "/"; cast1 = "(BAS_DBL)"; break;
        case (QBtoken)'^':    funame = "std::pow("; op=","; finish=")"; break;
        case MOD:             funame = "BasMODop("; op=","; finish=")"; break;
        default: assert(!"unexpected expr_binary token");
    }
    std::cout << fucast << funame << "("
              << cast0; CompileExpr(p0, subdefs);
    std::cout << ' ' << op << ' '
              << cast1; CompileExpr(p1, subdefs);
    std::cout << finish
              << ')';
}

void expr_userfun::Compile(ExprCompileDefs defs)
{
    std::string basename = BasUCASE(StripTypeSuffix(name));
    SubFuncP subp = FindSubFunc(basename);
    if(!subp)
    {
        fprintf(stderr, "Error: Sub/Func %s (%s) not found\n", name.c_str(), basename.c_str());
        return;
    }
    SubRoutine& sub = *subp;
    const std::vector<SubRoutine::ParamSpec>& paramdefs = sub.params;
    
    if(params.size() != paramdefs.size())
    {
        fprintf(stderr, "Error: Function %s is given %u arguments, expects only %u\n",
            name.c_str(), (unsigned)params.size(), (unsigned)paramdefs.size());
        return;
    }
    
    CompileFunctionCall(
        CPP_SUBFUNC_NAME(basename),
        params,
        paramdefs
    );
}

static void TranslateBuiltinFunc(
    QBtoken token,
    std::vector<exprp>& params,
    const std::string& pattern,
    ExprCompileDefs defs)
{
    const char* funcname = 0;
    #define tokcase(t, p) \
        case t: funcname = "B_Func"#t; goto DoParams;
    switch(token)
    {
#include "libdata-expr_numeric.inc"
#include "libdata-expr_string.inc"

        default: assert(!"unknown expr_builtin token");
    }
    return;
DoParams:
    std::vector<exprp> plist = params;
    std::vector<SubRoutine::ParamSpec> paramdefs
        = CreateParamDefsFromPattern(pattern, plist);
    CompileFunctionCall(
        funcname,
        plist,
        paramdefs
    );
    #undef tokcase
}

    
void expr_builtin_numeric::Compile(ExprCompileDefs defs)
{
    TranslateBuiltinFunc(token, params, pattern, defs);
}
void expr_builtin_string::Compile(ExprCompileDefs defs)
{
    TranslateBuiltinFunc(token, params, pattern, defs);
}

void expr_lubound::Compile(ExprCompileDefs defs)
{
    bool kind_u = token==UBOUND;
    
    const char* fname = kind_u?"UBOUND":"LBOUND";
    const char* bname = kind_u?"Ubound":"Lbound";
    
    expr_var* v = dynamic_cast<expr_var*>(&*var);
    if(!v) { ae:fprintf(stderr, "%s must only be used on arrays\n",fname); return; }
    if(!v->isarray()) goto ae;
    
    const std::string& name = v->getname();
    VarContext ctx = FindArrayType(name);
    if(!ctx) fprintf(stderr, "Cannot find definition of '%s'\n", name.c_str());
    TypeP type = ctx.info->type;
    
    switch(type->arraytype)
    {
        case TypeInfo::NotArray: goto ae;
        case TypeInfo::StaticArray:
        {
            if(expr_number* e = dynamic_cast<expr_number*>(&*num_index))
            {
                long index = e->get_i();
                if(index < 1 || index > type->dimensions.size())
                    fprintf(stderr, "Array subscript out of range\n");
                else
                    CompileExpr(new expr_number(
                        type->dimensions[index-1].first
                      + (kind_u ? type->dimensions[index-1].second-1 : 0) ));
            }
            else
            {
                std::cout << "({ int bound=0; switch(";
                CompileExpr(num_index, exprdef_cast_to_int);
                std::cout << ") {\n";
                for(unsigned a=0; a<type->dimensions.size(); ++a)
                    std::cout << "case " << (a+1) << ": bound="
                    << (type->dimensions[a].first
                      + (kind_u ? type->dimensions[a].second-1 : 0))
                    <<  "; break;\n"
                             "default: B_ERROR(9); \n" /* Subscript out of range */
                             "} bound; })";
            }
            break;
        }
        case TypeInfo::DynamicArray:
        case TypeInfo::ParamArray:
        {
            std::pair<std::string, std::string>
                base = GetTypeBaseNames(*type, BasUCASE(StripTypeSuffix(name)));

            std::cout << CPP_ARRAY_NAME(base.second)
                      << '.' << bname
                      << '(';
            CompileExpr(num_index, exprdef_cast_to_int);
            std::cout << ')';
            break;
        }
    }
}

// SPECIALS
void expr_label::Compile(ExprCompileDefs defs)    { assert(!"not reached"); }
void expr_caseelse::Compile(ExprCompileDefs defs) { assert(!"not reached"); }
void expr_caseis::Compile(ExprCompileDefs defs)   { assert(!"not reached"); }
void expr_caserange::Compile(ExprCompileDefs defs){ assert(!"not reached"); }
void expr_caseor::Compile(ExprCompileDefs defs)   { assert(!"not reached"); }

// LVALUES
void expr_var::Compile(ExprCompileDefs defs)
{
    TypeP basetype = type;
    if(defs != exprdef_membername)
    {
        VarContext ctx;

        if(isarray())
            ctx = FindArrayType(name);
        else
            ctx = FindScalarType(name);
        if(!ctx) fprintf(stderr, "Cannot find definition of '%s'\n", name.c_str());
        assert(ctx);
        assert(ctx.info);
        basetype = ctx.info->type;
    }
    assert(basetype);
    
    std::pair<std::string, std::string>
        base = GetTypeBaseNames(*basetype, BasUCASE(StripTypeSuffix(name)));

    if(defs==exprdef_membername)
        base.second = CPP_MEMBER_NAME(base.second);
    
    switch(basetype->arraytype)
    {
        case TypeInfo::NotArray:
            std::cout << base.second;
            break;
        case TypeInfo::StaticArray:
            std::cout << CPP_ARRAY_NAME(base.second);
            for(unsigned a=0; a<indexes.size(); ++a)
            {
                std::cout << '[';
                exprp index = indexes[a];
                if(basetype->dimensions[a].first != 0)
                {
                    index = new expr_binary
                        ((QBtoken)'-', index, new expr_number(basetype->dimensions[a].first));
                    index = index->optimized();
                }
                CompileExpr(index, exprdef_cast_to_int);
                std::cout << ']';
            }
            break;
        case TypeInfo::DynamicArray:
        case TypeInfo::ParamArray:
            std::cout << CPP_ARRAY_NAME(base.second);
            if(!array_as_whole)
            {
                std::cout << ".Index(" << indexes.size();
                for(unsigned a=0; a<indexes.size(); ++a)
                {
                    std::cout << ", ";
                    CompileExpr(indexes[a], exprdef_cast_to_int);
                }
                std::cout << ')';
            }
            break;
    }
    if(hasmember())
    {
        std::cout << '.';
        member->Compile(exprdef_membername); // not CompileExpr here
    }
}
void expr_returnvalueholder::Compile(ExprCompileDefs defs)
{
    std::cout << FunctionReturnVarName;
}
